"""invrs_opt - Optimization algorithms for inverse design.

Copyright (c) 2023 The INVRS-IO authors.
"""

__version__ = "v0.7.1"
__author__ = "Martin F. Schubert <mfschubert@gmail.com>"

from invrs_opt import parameterization as parameterization

from invrs_opt.optimizers.lbfgsb import (
    density_lbfgsb as density_lbfgsb,
    lbfgsb as lbfgsb,
    levelset_lbfgsb as levelset_lbfgsb,
)

from invrs_opt.optimizers.wrapped_optax import (
    density_wrapped_optax as density_wrapped_optax,
    levelset_wrapped_optax as levelset_wrapped_optax,
    wrapped_optax as wrapped_optax,
)



"""Base types for density parameterizations.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
from typing import Any, Optional, Protocol, Sequence, Tuple

import jax.numpy as jnp
import numpy as onp
from jax import tree_util
from totypes import json_utils, types

Array = jnp.ndarray | onp.ndarray[Any, Any]
PyTree = Any


class ParameterizedDensity2DArrayBase:
    """Base class for parameterized density arrays."""

    pass


class FromDensityFn(Protocol):
    """Generate the latent representation of a density array."""

    def __call__(
        self, density: types.Density2DArray
    ) -> ParameterizedDensity2DArrayBase:
        ...


class ToDensityFn(Protocol):
    """Generate a density from its latent representation."""

    def __call__(self, params: PyTree) -> types.Density2DArray:
        ...


class ConstraintsFn(Protocol):
    """Compute constraints for a latent representation of a density array."""

    def __call__(self, params: PyTree) -> jnp.ndarray:
        ...


@dataclasses.dataclass
class Density2DParameterization:
    """Stores `(from_density, to_density, constraints)` function triple."""

    from_density: FromDensityFn
    to_density: ToDensityFn
    constraints: ConstraintsFn


@dataclasses.dataclass
class Density2DMetadata:
    """Stores the metadata of a `Density2DArray`."""

    lower_bound: float
    upper_bound: float
    fixed_solid: Optional[Array]
    fixed_void: Optional[Array]
    minimum_width: int
    minimum_spacing: int
    periodic: Sequence[bool]
    symmetries: Sequence[str]

    def __post_init__(self) -> None:
        self.periodic = tuple(self.periodic)
        self.symmetries = tuple(self.symmetries)


def _flatten_density_2d_metadata(
    metadata: Density2DMetadata,
) -> Tuple[
    Tuple[()],
    Tuple[
        float,
        float,
        types.HashableWrapper,
        types.HashableWrapper,
        int,
        int,
        Sequence[bool],
        Sequence[str],
    ],
]:
    """Flattens a `Density2DMetadata` into children and auxilliary data."""
    return (
        (),
        (
            metadata.lower_bound,
            metadata.upper_bound,
            types.HashableWrapper(metadata.fixed_solid),
            types.HashableWrapper(metadata.fixed_void),
            metadata.minimum_width,
            metadata.minimum_spacing,
            metadata.periodic,
            metadata.symmetries,
        ),
    )


def _unflatten_density_2d_metadata(
    aux: Tuple[
        float,
        float,
        types.HashableWrapper,
        types.HashableWrapper,
        int,
        int,
        Sequence[bool],
        Sequence[str],
    ],
    children: Tuple[()],
) -> Density2DMetadata:
    """Unflattens a flattened `Density2DMetadata`."""
    del children
    (
        lower_bound,
        upper_bound,
        wrapped_fixed_solid,
        wrapped_fixed_void,
        minimum_width,
        minimum_spacing,
        periodic,
        symmetries,
    ) = aux
    return Density2DMetadata(
        lower_bound=lower_bound,
        upper_bound=upper_bound,
        fixed_solid=wrapped_fixed_solid.array,  # type: ignore[arg-type]
        fixed_void=wrapped_fixed_void.array,  # type: ignore[arg-type]
        minimum_width=minimum_width,
        minimum_spacing=minimum_spacing,
        periodic=tuple(periodic),
        symmetries=tuple(symmetries),
    )


tree_util.register_pytree_node(
    Density2DMetadata,
    flatten_func=_flatten_density_2d_metadata,
    unflatten_func=_unflatten_density_2d_metadata,
)

json_utils.register_custom_type(Density2DMetadata)



"""Defines a jax-style wrapper for scipy's L-BFGS-B algorithm.

Copyright (c) 2023 The INVRS-IO authors.
"""

import copy
import dataclasses
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as onp
import optax  # type: ignore[import-untyped]
from jax import flatten_util, tree_util
from scipy.optimize._lbfgsb_py import (  # type: ignore[import-untyped]
    _lbfgsb as scipy_lbfgsb,
)
from totypes import types

from invrs_opt.optimizers import base
from invrs_opt.parameterization import (
    base as parameterization_base,
    filter_project,
    gaussian_levelset,
    pixel,
)

NDArray = onp.ndarray[Any, Any]
PyTree = Any
ElementwiseBound = Union[NDArray, Sequence[Optional[float]]]
JaxLbfgsbDict = Dict[str, jnp.ndarray]
LbfgsbState = Tuple[PyTree, PyTree, JaxLbfgsbDict]


# Task message prefixes for the underlying L-BFGS-B implementation.
TASK_START = b"START"
TASK_FG = b"FG"
TASK_CONVERGED = b"CONVERGENCE"

UPDATE_IPRINT = -1

# Maximum value for the `maxcor` parameter in the L-BFGS-B scheme.
MAXCOR_MAX_VALUE = 100
DEFAULT_MAXCOR = 20
DEFAULT_LINE_SEARCH_MAX_STEPS = 100
DEFAULT_FTOL = 0.0
DEFAULT_GTOL = 0.0

# Maps bound scenarios to integers.
BOUNDS_MAP: Dict[Tuple[bool, bool], int] = {
    (True, True): 0,  # Both upper and lower bound are `None`.
    (False, True): 1,  # Only upper bound is `None`.
    (False, False): 2,  # Neither of the bounds are `None`.
    (True, False): 3,  # Only the lower bound is `None`.
}

FORTRAN_INT = scipy_lbfgsb.types.intvar.dtype


def lbfgsb(
    *,
    maxcor: int = DEFAULT_MAXCOR,
    line_search_max_steps: int = DEFAULT_LINE_SEARCH_MAX_STEPS,
    ftol: float = DEFAULT_FTOL,
    gtol: float = DEFAULT_GTOL,
) -> base.Optimizer:
    """Optimizer implementing the standard L-BFGS-B algorithm.

    The standard L-BFGS-B algorithm uses the direct pixel parameterization for density
    arrays, which simply enforces that values are between the declared upper and lower
    bounds of the density.

    When an optimization is determined to have converged (by `ftol` or `gtol` criteria)
    the optimizer `params` function will simply return the optimal parameters. The
    convergence can be queried by `is_converged(state)`.

    Args:
        maxcor: The maximum number of variable metric corrections used to define the
            limited memory matrix, in the L-BFGS-B scheme.
        line_search_max_steps: The maximum number of steps in the line search.
        ftol: Convergence criteria based on function values. See scipy documentation
            for details.
        gtol: Convergence criteria based on gradient.

    Returns:
        The `Optimizer` implementing the L-BFGS-B optimizer.
    """
    return parameterized_lbfgsb(
        density_parameterization=None,
        penalty=0.0,
        maxcor=maxcor,
        line_search_max_steps=line_search_max_steps,
        ftol=ftol,
        gtol=gtol,
    )


def density_lbfgsb(
    *,
    beta: float,
    maxcor: int = DEFAULT_MAXCOR,
    line_search_max_steps: int = DEFAULT_LINE_SEARCH_MAX_STEPS,
    ftol: float = DEFAULT_FTOL,
    gtol: float = DEFAULT_GTOL,
) -> base.Optimizer:
    """Optimizer using L-BFGS-B algorithm with filter-project density parameterization.

    In the filter-project density parameterization, the optimization variable
    associated with a density array is a latent density array; the density is obtained
    by convolving (i.e. "filtering") the latent density with a Gaussian kernel having
    full-width at half-maximum equal to the length scale (the mean of declared minimum
    width and minimum spacing). Then, a tanh nonlinearity is used as a smooth threshold
    operation ("projection").

    When an optimization is determined to have converged (by `ftol` or `gtol` criteria)
    the optimizer `params` function will simply return the optimal parameters. The
    convergence can be queried by `is_converged(state)`.

    Args:
        beta: Determines the sharpness of the thresholding operation.
        maxcor: The maximum number of variable metric corrections used to define the
            limited memory matrix, in the L-BFGS-B scheme.
        line_search_max_steps: The maximum number of steps in the line search.
        ftol: Convergence criteria based on function values. See scipy documentation
            for details.
        gtol: Convergence criteria based on gradient.

    Returns:
        The `Optimizer` implementing the L-BFGS-B optimizer.
    """
    return parameterized_lbfgsb(
        density_parameterization=filter_project.filter_project(beta=beta),
        penalty=0.0,
        maxcor=maxcor,
        line_search_max_steps=line_search_max_steps,
        ftol=ftol,
        gtol=gtol,
    )


def levelset_lbfgsb(
    *,
    penalty: float,
    length_scale_spacing_factor: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_SPACING_FACTOR
    ),
    length_scale_fwhm_factor: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_FWHM_FACTOR
    ),
    length_scale_constraint_factor: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_CONSTRAINT_FACTOR
    ),
    smoothing_factor: int = gaussian_levelset.DEFAULT_SMOOTHING_FACTOR,
    length_scale_constraint_beta: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_CONSTRAINT_BETA
    ),
    length_scale_constraint_weight: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_CONSTRAINT_WEIGHT
    ),
    curvature_constraint_weight: float = (
        gaussian_levelset.DEFAULT_CURVATURE_CONSTRAINT_WEIGHT
    ),
    fixed_pixel_constraint_weight: float = (
        gaussian_levelset.DEFAULT_FIXED_PIXEL_CONSTRAINT_WEIGHT
    ),
    init_optimizer: optax.GradientTransformation = (
        gaussian_levelset.DEFAULT_INIT_OPTIMIZER
    ),
    init_steps: int = gaussian_levelset.DEFAULT_INIT_STEPS,
    maxcor: int = DEFAULT_MAXCOR,
    line_search_max_steps: int = DEFAULT_LINE_SEARCH_MAX_STEPS,
    ftol: float = DEFAULT_FTOL,
    gtol: float = DEFAULT_GTOL,
) -> base.Optimizer:
    """Optimizer using L-BFGS-B algorithm with levelset density parameterization.

    In the levelset parameterization, the optimization variable associated with a
    density array is an array giving the amplitudes of Gaussian radial basis functions
    that represent a levelset function over the domain of the density. In the levelset
    parameterization, gradients are nonzero only at the edges of features, and in
    general the topology of a solution does not change during the course of
    optimization.

    The spacing and full-width at half-maximum of the Gaussian basis functions gives
    some amount of control over length scales. In addition, constraints associated with
    length scale, radius of curvature, and deviation from fixed pixels are
    automatically computed and penalized with a weight given by `penalty`. In general,
    this helps ensure that features in an optimized density array violate the specified
    constraints to a lesser degree. The constraints are based on "Analytical level set
    fabrication constraints for inverse design," by D. Vercruysse et al. (2019).

    When an optimization is determined to have converged (by `ftol` or `gtol` criteria)
    the optimizer `params` function will simply return the optimal parameters. The
    convergence can be queried by `is_converged(state)`.

    Args:
        penalty: The weight of the fabrication penalty, which combines length scale,
            curvature, and fixed pixel constraints.
        length_scale_spacing_factor: The number of levelset control points per unit of
            minimum length scale (mean of density minimum width and minimum spacing).
        length_scale_fwhm_factor: The ratio of Gaussian full-width at half-maximum to
            the minimum length scale.
        length_scale_constraint_factor: Multiplies the target length scale in the
            levelset constraints. A value greater than 1 is pessimistic and drives the
            solution to have a larger length scale (relative to smaller values).
        smoothing_factor: For values greater than 1, the density is initially computed
            at higher resolution and then downsampled, yielding smoother geometries.
        length_scale_constraint_beta: Controls relaxation of the length scale
            constraint near the zero level.
        length_scale_constraint_weight: The weight of the length scale constraint in
            the overall fabrication constraint peenalty.
        curvature_constraint_weight: The weight of the curvature constraint.
        fixed_pixel_constraint_weight: The weight of the fixed pixel constraint.
        init_optimizer: The optimizer used in the initialization of the levelset
            parameterization. At initialization, the latent parameters are optimized so
            that the initial parameters match the binarized initial density.
        init_steps: The number of optimization steps used in the initialization.
        maxcor: The maximum number of variable metric corrections used to define the
            limited memory matrix, in the L-BFGS-B scheme.
        line_search_max_steps: The maximum number of steps in the line search.
        ftol: Convergence criteria based on function values. See scipy documentation
            for details.
        gtol: Convergence criteria based on gradient.

    Returns:
        The `Optimizer` implementing the L-BFGS-B optimizer.
    """
    return parameterized_lbfgsb(
        density_parameterization=gaussian_levelset.gaussian_levelset(
            length_scale_spacing_factor=length_scale_spacing_factor,
            length_scale_fwhm_factor=length_scale_fwhm_factor,
            length_scale_constraint_factor=length_scale_constraint_factor,
            smoothing_factor=smoothing_factor,
            length_scale_constraint_beta=length_scale_constraint_beta,
            length_scale_constraint_weight=length_scale_constraint_weight,
            curvature_constraint_weight=curvature_constraint_weight,
            fixed_pixel_constraint_weight=fixed_pixel_constraint_weight,
            init_optimizer=init_optimizer,
            init_steps=init_steps,
        ),
        penalty=penalty,
        maxcor=maxcor,
        line_search_max_steps=line_search_max_steps,
        ftol=ftol,
        gtol=gtol,
    )


# -----------------------------------------------------------------------------
# Base parameterized L-BFGS-B optimizer.
# -----------------------------------------------------------------------------


def parameterized_lbfgsb(
    density_parameterization: Optional[parameterization_base.Density2DParameterization],
    penalty: float,
    maxcor: int = DEFAULT_MAXCOR,
    line_search_max_steps: int = DEFAULT_LINE_SEARCH_MAX_STEPS,
    ftol: float = DEFAULT_FTOL,
    gtol: float = DEFAULT_GTOL,
) -> base.Optimizer:
    """Optimizer using L-BFGS-B optimizer with specified density parameterization.

    This optimizer wraps scipy's implementation of the algorithm, and provides
    a jax-style API to the scheme. The optimizer works with custom types such
    as the `BoundedArray` to constrain the optimization variable.

    Args:
        density_parameterization: The parameterization to be used, or `None`. When no
            parameterization is given, the direct pixel parameterization is used for
            density arrays.
        penalty: The weight of the scalar penalty formed from the constraints of the
            parameterization.
        maxcor: The maximum number of variable metric corrections used to define the
            limited memory matrix, in the L-BFGS-B scheme.
        line_search_max_steps: The maximum number of steps in the line search.
        ftol: Convergence criteria based on function values. See scipy documentation
            for details.
        gtol: Convergence criteria based on gradient.

    Returns:
        The `base.Optimizer`.
    """
    if not isinstance(maxcor, int) or maxcor < 1 or maxcor > MAXCOR_MAX_VALUE:
        raise ValueError(
            f"`maxcor` must be greater than 0 and less than "
            f"{MAXCOR_MAX_VALUE}, but got {maxcor}"
        )

    if not isinstance(line_search_max_steps, int) or line_search_max_steps < 1:
        raise ValueError(
            f"`line_search_max_steps` must be greater than 0 but got "
            f"{line_search_max_steps}"
        )

    if density_parameterization is None:
        density_parameterization = pixel.pixel()

    def _init_latents(params: PyTree) -> PyTree:
        def _leaf_init_latents(leaf: Any) -> Any:
            leaf = _clip(leaf)
            if not _is_density(leaf) or density_parameterization is None:
                return leaf
            return density_parameterization.from_density(leaf)

        return tree_util.tree_map(_leaf_init_latents, params, is_leaf=_is_custom_type)

    def _params_from_latents(latent_params: PyTree) -> PyTree:
        def _leaf_params_from_latents(leaf: Any) -> Any:
            if not _is_parameterized_density(leaf) or density_parameterization is None:
                return leaf
            return density_parameterization.to_density(leaf)

        return tree_util.tree_map(
            _leaf_params_from_latents,
            latent_params,
            is_leaf=_is_parameterized_density,
        )

    def _constraint_loss(latent_params: PyTree) -> jnp.ndarray:
        def _constraint_loss_leaf(
            params: parameterization_base.ParameterizedDensity2DArrayBase,
        ) -> jnp.ndarray:
            constraints = density_parameterization.constraints(params)
            constraints = tree_util.tree_map(
                lambda x: jnp.sum(jnp.maximum(x, 0.0)),
                constraints,
            )
            return jnp.sum(jnp.asarray(constraints))

        losses = [0.0] + [
            _constraint_loss_leaf(p)
            for p in tree_util.tree_leaves(
                latent_params, is_leaf=_is_parameterized_density
            )
            if _is_parameterized_density(p)
        ]
        return penalty * jnp.sum(jnp.asarray(losses))

    def init_fn(params: PyTree) -> LbfgsbState:
        """Initializes the optimization state."""

        def _init_state_pure(latent_params: PyTree) -> Tuple[PyTree, JaxLbfgsbDict]:
            lower_bound = types.extract_lower_bound(latent_params)
            upper_bound = types.extract_upper_bound(latent_params)
            scipy_lbfgsb_state = ScipyLbfgsbState.init(
                x0=_to_numpy(latent_params),
                lower_bound=_bound_for_params(lower_bound, latent_params),
                upper_bound=_bound_for_params(upper_bound, latent_params),
                maxcor=maxcor,
                line_search_max_steps=line_search_max_steps,
                ftol=ftol,
                gtol=gtol,
            )
            latent_params = _to_pytree(scipy_lbfgsb_state.x, latent_params)
            return latent_params, scipy_lbfgsb_state.to_jax()

        latent_params = _init_latents(params)
        latent_params, jax_lbfgsb_state = jax.pure_callback(
            _init_state_pure, _example_state(latent_params, maxcor), latent_params
        )
        return _params_from_latents(latent_params), latent_params, jax_lbfgsb_state

    def params_fn(state: LbfgsbState) -> PyTree:
        """Returns the parameters for the given `state`."""
        params, _, _ = state
        return params

    def update_fn(
        *,
        grad: PyTree,
        value: float,
        params: PyTree,
        state: LbfgsbState,
    ) -> LbfgsbState:
        """Updates the state."""
        del params

        def _update_pure(
            flat_latent_grad: PyTree,
            value: jnp.ndarray,
            jax_lbfgsb_state: JaxLbfgsbDict,
        ) -> Tuple[PyTree, JaxLbfgsbDict]:
            assert onp.size(value) == 1
            scipy_lbfgsb_state = ScipyLbfgsbState.from_jax(jax_lbfgsb_state)
            scipy_lbfgsb_state.update(
                grad=onp.array(flat_latent_grad, dtype=onp.float64),
                value=onp.array(value, dtype=onp.float64),
            )
            flat_latent_params = jnp.asarray(scipy_lbfgsb_state.x)
            return flat_latent_params, scipy_lbfgsb_state.to_jax()

        _, latent_params, jax_lbfgsb_state = state
        _, vjp_fn = jax.vjp(_params_from_latents, latent_params)
        (latent_grad,) = vjp_fn(grad)

        if not (
            tree_util.tree_structure(latent_grad)
            == tree_util.tree_structure(latent_params)  # type: ignore[operator]
        ):
            raise ValueError(
                f"Tree structure of `latent_grad` was different than expected, got \n"
                f"{tree_util.tree_structure(latent_grad)} but expected \n"
                f"{tree_util.tree_structure(latent_params)}."
            )

        (
            constraint_loss_value,
            constraint_loss_grad,
        ) = jax.value_and_grad(
            _constraint_loss
        )(latent_params)
        value += constraint_loss_value
        latent_grad = tree_util.tree_map(
            lambda a, b: a + b, latent_grad, constraint_loss_grad
        )

        flat_latent_grad, unflatten_fn = flatten_util.ravel_pytree(
            latent_grad
        )  # type: ignore[no-untyped-call]

        flat_latent_params, jax_lbfgsb_state = jax.pure_callback(
            _update_pure,
            (flat_latent_grad, jax_lbfgsb_state),
            flat_latent_grad,
            value,
            jax_lbfgsb_state,
        )
        latent_params = unflatten_fn(flat_latent_params)
        return _params_from_latents(latent_params), latent_params, jax_lbfgsb_state

    return base.Optimizer(
        init=init_fn,
        params=params_fn,
        update=update_fn,
    )


def is_converged(state: LbfgsbState) -> jnp.ndarray:
    """Returns `True` if the optimization has converged."""
    return state[2]["converged"]


# ------------------------------------------------------------------------------
# Helper functions.
# ------------------------------------------------------------------------------


def _is_density(leaf: Any) -> Any:
    """Return `True` if `leaf` is a density array."""
    return isinstance(leaf, types.Density2DArray)


def _is_parameterized_density(leaf: Any) -> Any:
    """Return `True` if `leaf` is a parameterized density array."""
    return isinstance(leaf, parameterization_base.ParameterizedDensity2DArrayBase)


def _is_custom_type(leaf: Any) -> bool:
    """Return `True` if `leaf` is a recognized custom type."""
    return isinstance(leaf, (types.BoundedArray, types.Density2DArray))


def _clip(pytree: PyTree) -> PyTree:
    """Clips leaves on `pytree` to their bounds."""

    def _clip_fn(leaf: Any) -> Any:
        if not _is_custom_type(leaf):
            return leaf
        if leaf.lower_bound is None and leaf.upper_bound is None:
            return leaf
        return tree_util.tree_map(
            lambda x: jnp.clip(x, leaf.lower_bound, leaf.upper_bound), leaf
        )

    return tree_util.tree_map(_clip_fn, pytree, is_leaf=_is_custom_type)


def _to_numpy(params: PyTree) -> NDArray:
    """Flattens a `params` pytree into a single rank-1 numpy array."""
    x, _ = flatten_util.ravel_pytree(params)  # type: ignore[no-untyped-call]
    return onp.asarray(x, dtype=onp.float64)


def _to_pytree(x_flat: NDArray, params: PyTree) -> PyTree:
    """Restores a pytree from a flat numpy array using the structure of `params`.

    Note that the returned pytree includes jax array leaves.

    Args:
        x_flat: The rank-1 numpy array to be restored.
        params: A pytree of parameters whose structure is replicated in the restored
            pytree.

    Returns:
        The restored pytree, with jax array leaves.
    """
    _, unflatten_fn = flatten_util.ravel_pytree(params)  # type: ignore[no-untyped-call]
    return unflatten_fn(jnp.asarray(x_flat, dtype=float))


def _bound_for_params(bound: PyTree, params: PyTree) -> ElementwiseBound:
    """Generates a bound vector for the `params`.

    The `bound` can be specified in various ways; it may be `None` or a scalar,
    which then applies to all arrays in `params`. It may be a pytree with
    structure matching that of `params`, where each leaf is either `None`, a
    scalar, or an array matching the shape of the corresponding leaf in `params`.

    The returned bound is a flat array suitable for use with `ScipyLbfgsbState`.

    Args:
        bound: The pytree of bounds.
        params: The pytree of parameters.

    Returns:
        The flat elementwise bound.
    """

    if bound is None or onp.isscalar(bound):
        bound = tree_util.tree_map(
            lambda _: bound,
            params,
            is_leaf=lambda x: isinstance(x, types.CUSTOM_TYPES),
        )

    bound_leaves, bound_treedef = tree_util.tree_flatten(
        bound, is_leaf=lambda x: x is None
    )
    params_leaves = tree_util.tree_leaves(params, is_leaf=lambda x: x is None)

    # `bound` should be a pytree of arrays or `None`, while `params` may
    # include custom pytree nodes. Convert the custom nodes into standard
    # types to facilitate validation that the tree structures match.
    params_treedef = tree_util.tree_structure(
        tree_util.tree_map(
            lambda x: 0.0,
            tree=params,
            is_leaf=lambda x: x is None or isinstance(x, types.CUSTOM_TYPES),
        )
    )
    if bound_treedef != params_treedef:  # type: ignore[operator]
        raise ValueError(
            f"Tree structure of `bound` and `params` must match, but got "
            f"{bound_treedef} and {params_treedef}, respectively."
        )

    bound_flat = []
    for b, p in zip(bound_leaves, params_leaves):
        if p is None:
            continue
        if b is None or onp.isscalar(b) or onp.shape(b) == ():
            bound_flat += [b] * onp.size(p)
        else:
            if b.shape != p.shape:
                raise ValueError(
                    f"`bound` must be `None`, a scalar, or have shape matching "
                    f"`params`, but got shape {b.shape} when params has shape "
                    f"{p.shape}."
                )
            bound_flat += b.flatten().tolist()

    return bound_flat


def _example_state(params: PyTree, maxcor: int) -> PyTree:
    """Return an example state for the given `params` and `maxcor`."""
    params_flat, _ = flatten_util.ravel_pytree(params)  # type: ignore[no-untyped-call]
    n = params_flat.size
    float_params = tree_util.tree_map(lambda x: jnp.asarray(x, dtype=float), params)
    example_jax_lbfgsb_state = dict(
        x=jnp.zeros(n, dtype=float),
        converged=jnp.asarray(False),
        _maxcor=jnp.zeros((), dtype=int),
        _line_search_max_steps=jnp.zeros((), dtype=int),
        _ftol=jnp.zeros((), dtype=float),
        _gtol=jnp.zeros((), dtype=float),
        _wa=jnp.ones(_wa_size(n=n, maxcor=maxcor), dtype=float),
        _iwa=jnp.ones(n * 3, dtype=jnp.int32),  # Fortran int
        _task=jnp.zeros(59, dtype=int),
        _csave=jnp.zeros(59, dtype=int),
        _lsave=jnp.zeros(4, dtype=jnp.int32),  # Fortran int
        _isave=jnp.zeros(44, dtype=jnp.int32),  # Fortran int
        _dsave=jnp.zeros(29, dtype=float),
        _lower_bound=jnp.zeros(n, dtype=float),
        _upper_bound=jnp.zeros(n, dtype=float),
        _bound_type=jnp.zeros(n, dtype=int),
    )
    return float_params, example_jax_lbfgsb_state


# ------------------------------------------------------------------------------
# Wrapper for scipy's L-BFGS-B implementation.
# ------------------------------------------------------------------------------


@dataclasses.dataclass
class ScipyLbfgsbState:
    """Stores the state of a scipy L-BFGS-B minimization.

    This class enables optimization with a more functional style, giving the user
    control over the optimization loop. Example usage is as follows:

        value_fn = lambda x: onp.sum(x**2)
        grad_fn = lambda x: 2 * x

        x0 = onp.asarray([0.1, 0.2, 0.3])
        lb = [None, -1, 0.1]
        ub = [None, None, None]
        state = ScipyLbfgsbState.init(
            x0=x0, lower_bound=lb, upper_bound=ub, maxcor=20
        )

        for _ in range(10):
            value = value_fn(state.x)
            grad = grad_fn(state.x)
            state.update(grad, value)

    This example converges with `state.x` equal to `(0, 0, 0.1)` and value equal
    to `0.01`.

    Attributes:
        x: The current solution vector.
    """

    x: NDArray
    converged: NDArray
    # Private attributes correspond to internal variables in the `scipy.optimize.
    # lbfgsb._minimize_lbfgsb` function.
    _maxcor: int
    _line_search_max_steps: int
    _ftol: NDArray
    _gtol: NDArray
    _wa: NDArray
    _iwa: NDArray
    _task: NDArray
    _csave: NDArray
    _lsave: NDArray
    _isave: NDArray
    _dsave: NDArray
    _lower_bound: NDArray
    _upper_bound: NDArray
    _bound_type: NDArray

    def __post_init__(self) -> None:
        """Validates the datatypes for all state attributes."""
        _validate_array_dtype(self.x, onp.float64)
        _validate_array_dtype(self._wa, onp.float64)
        _validate_array_dtype(self._iwa, FORTRAN_INT)
        _validate_array_dtype(self._task, "S60")
        _validate_array_dtype(self._csave, "S60")
        _validate_array_dtype(self._lsave, FORTRAN_INT)
        _validate_array_dtype(self._isave, FORTRAN_INT)
        _validate_array_dtype(self._dsave, onp.float64)
        _validate_array_dtype(self._lower_bound, onp.float64)
        _validate_array_dtype(self._upper_bound, onp.float64)
        _validate_array_dtype(self._bound_type, int)

    def to_jax(self) -> Dict[str, jnp.ndarray]:
        """Generates a dictionary of jax arrays defining the state."""
        return dict(
            x=jnp.asarray(self.x),
            converged=jnp.asarray(self.converged),
            _maxcor=jnp.asarray(self._maxcor),
            _line_search_max_steps=jnp.asarray(self._line_search_max_steps),
            _ftol=jnp.asarray(self._ftol),
            _gtol=jnp.asarray(self._gtol),
            _wa=jnp.asarray(self._wa),
            _iwa=jnp.asarray(self._iwa),
            _task=_array_from_s60_str(self._task),
            _csave=_array_from_s60_str(self._csave),
            _lsave=jnp.asarray(self._lsave),
            _isave=jnp.asarray(self._isave),
            _dsave=jnp.asarray(self._dsave),
            _lower_bound=jnp.asarray(self._lower_bound),
            _upper_bound=jnp.asarray(self._upper_bound),
            _bound_type=jnp.asarray(self._bound_type),
        )

    @classmethod
    def from_jax(cls, state_dict: Dict[str, jnp.ndarray]) -> "ScipyLbfgsbState":
        """Converts a dictionary of jax arrays to a `ScipyLbfgsbState`."""
        state_dict = copy.deepcopy(state_dict)
        return ScipyLbfgsbState(
            x=onp.array(state_dict["x"], dtype=onp.float64),
            converged=onp.asarray(state_dict["converged"], dtype=bool),
            _maxcor=int(state_dict["_maxcor"]),
            _line_search_max_steps=int(state_dict["_line_search_max_steps"]),
            _ftol=onp.asarray(state_dict["_ftol"], dtype=onp.float64),
            _gtol=onp.asarray(state_dict["_gtol"], dtype=onp.float64),
            _wa=onp.array(state_dict["_wa"], onp.float64),
            _iwa=onp.array(state_dict["_iwa"], dtype=FORTRAN_INT),
            _task=_s60_str_from_array(state_dict["_task"]),
            _csave=_s60_str_from_array(state_dict["_csave"]),
            _lsave=onp.array(state_dict["_lsave"], dtype=FORTRAN_INT),
            _isave=onp.array(state_dict["_isave"], dtype=FORTRAN_INT),
            _dsave=onp.array(state_dict["_dsave"], dtype=onp.float64),
            _lower_bound=onp.asarray(state_dict["_lower_bound"], dtype=onp.float64),
            _upper_bound=onp.asarray(state_dict["_upper_bound"], dtype=onp.float64),
            _bound_type=onp.asarray(state_dict["_bound_type"], dtype=int),
        )

    @classmethod
    def init(
        cls,
        x0: NDArray,
        lower_bound: ElementwiseBound,
        upper_bound: ElementwiseBound,
        maxcor: int,
        line_search_max_steps: int,
        ftol: float,
        gtol: float,
    ) -> "ScipyLbfgsbState":
        """Initializes the `ScipyLbfgsbState` for `x0`.

        Args:
            x0: Array giving the initial solution vector.
            lower_bound: Array giving the elementwise optional lower bound.
            upper_bound: Array giving the elementwise optional upper bound.
            maxcor: The maximum number of variable metric corrections used to define
                the limited memory matrix, in the L-BFGS-B scheme.
            line_search_max_steps: The maximum number of steps in the line search.
            ftol: Tolerance for stopping criteria based on function values. See scipy
                documentation for details.
            gtol: Tolerance for stopping criteria based on gradient.

        Returns:
            The `ScipyLbfgsbState`.
        """
        x0 = onp.asarray(x0)
        if x0.ndim > 1:
            raise ValueError(f"`x0` must be rank-1 but got shape {x0.shape}.")
        lower_bound = onp.asarray(lower_bound)
        upper_bound = onp.asarray(upper_bound)
        if x0.shape != lower_bound.shape or x0.shape != upper_bound.shape:
            raise ValueError(
                f"`x0`, `lower_bound`, and `upper_bound` must have matching "
                f"shape but got shapes {x0.shape}, {lower_bound.shape}, and "
                f"{upper_bound.shape}, respectively."
            )
        if maxcor < 1:
            raise ValueError(f"`maxcor` must be positive but got {maxcor}.")

        n = x0.size
        lower_bound_array, upper_bound_array, bound_type = _configure_bounds(
            lower_bound, upper_bound
        )
        task = onp.zeros(1, "S60")
        task[:] = TASK_START

        # See initialization of internal variables in the `lbfgsb._minimize_lbfgsb`
        # function.
        wa_size = _wa_size(n=n, maxcor=maxcor)
        state = ScipyLbfgsbState(
            x=onp.array(x0, onp.float64),
            converged=onp.asarray(False),
            _maxcor=maxcor,
            _line_search_max_steps=line_search_max_steps,
            _ftol=onp.asarray(ftol, onp.float64),
            _gtol=onp.asarray(gtol, onp.float64),
            _wa=onp.zeros(wa_size, onp.float64),
            _iwa=onp.zeros(3 * n, FORTRAN_INT),
            _task=task,
            _csave=onp.zeros(1, "S60"),
            _lsave=onp.zeros(4, FORTRAN_INT),
            _isave=onp.zeros(44, FORTRAN_INT),
            _dsave=onp.zeros(29, onp.float64),
            _lower_bound=lower_bound_array,
            _upper_bound=upper_bound_array,
            _bound_type=bound_type,
        )
        # The initial state requires an update with zero value and gradient. This
        # is because the initial task is "START", which does not actually require
        # value and gradient evaluation.
        state.update(onp.zeros(x0.shape, onp.float64), onp.zeros((), onp.float64))
        return state

    def update(
        self,
        grad: NDArray,
        value: NDArray,
    ) -> None:
        """Performs an in-place update of the `ScipyLbfgsbState` if not converged.

        Args:
            grad: The function gradient for the current `x`.
            value: The scalar function value for the current `x`.
        """
        if self.converged:
            return
        if grad.shape != self.x.shape:
            raise ValueError(
                f"`grad` must have the same shape as attribute `x`, but got shapes "
                f"{grad.shape} and {self.x.shape}, respectively."
            )
        if value.shape != ():
            raise ValueError(f"`value` must be a scalar but got shape {value.shape}.")

        # The `setulb` function will sometimes return with a task that does not
        # require a value and gradient evaluation. In this case we simply call it
        # again, advancing past such "dummy" steps.
        for _ in range(3):
            scipy_lbfgsb.setulb(
                m=self._maxcor,
                x=self.x,
                l=self._lower_bound,
                u=self._upper_bound,
                nbd=self._bound_type,
                f=value,
                g=grad,
                factr=self._ftol / onp.finfo(float).eps,
                pgtol=self._gtol,
                wa=self._wa,
                iwa=self._iwa,
                task=self._task,
                iprint=UPDATE_IPRINT,
                csave=self._csave,
                lsave=self._lsave,
                isave=self._isave,
                dsave=self._dsave,
                maxls=self._line_search_max_steps,
            )
            task_str = self._task.tobytes()
            if task_str.startswith(TASK_CONVERGED):
                self.converged = onp.asarray(True)
            if task_str.startswith(TASK_FG):
                break


def _wa_size(n: int, maxcor: int) -> int:
    """Return the size of the `wa` attribute of lbfgsb state."""
    return 2 * maxcor * n + 5 * n + 11 * maxcor**2 + 8 * maxcor


def _validate_array_dtype(x: NDArray, dtype: Union[type, str]) -> None:
    """Validates that `x` is an array with the specified `dtype`."""
    if not isinstance(x, onp.ndarray):
        raise ValueError(f"`x` must be an `onp.ndarray` but got {type(x)}")
    if x.dtype != dtype:
        raise ValueError(f"`x` must have dtype {dtype} but got {x.dtype}")


def _configure_bounds(
    lower_bound: ElementwiseBound,
    upper_bound: ElementwiseBound,
) -> Tuple[NDArray, NDArray, NDArray]:
    """Configures the bounds for an L-BFGS-B optimization."""
    bound_type = [
        BOUNDS_MAP[(lower is None, upper is None)]
        for lower, upper in zip(lower_bound, upper_bound)
    ]
    lower_bound_array = [0.0 if x is None else x for x in lower_bound]
    upper_bound_array = [0.0 if x is None else x for x in upper_bound]
    return (
        onp.asarray(lower_bound_array, onp.float64),
        onp.asarray(upper_bound_array, onp.float64),
        onp.asarray(bound_type),
    )


def _array_from_s60_str(s60_str: NDArray) -> jnp.ndarray:
    """Return a jax array for a numpy s60 string."""
    assert s60_str.shape == (1,)
    chars = [int(o) for o in s60_str[0]]
    chars.extend([32] * (59 - len(chars)))
    return jnp.asarray(chars, dtype=int)


def _s60_str_from_array(array: jnp.ndarray) -> NDArray:
    """Return a numpy s60 string for a jax array."""
    return onp.asarray(
        [b"".join(int(i).to_bytes(length=1, byteorder="big") for i in array)],
        dtype="S60",
    )



"""Defines the direct pixel parameterization for density arrays.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses

import jax.numpy as jnp
from jax import tree_util
from totypes import json_utils, types

from invrs_opt.parameterization import base


@dataclasses.dataclass
class PixelParams(base.ParameterizedDensity2DArrayBase):
    """Stores latent parameters of the direct pixel parameterization."""

    density: types.Density2DArray


tree_util.register_dataclass(PixelParams, data_fields=["density"], meta_fields=[])


json_utils.register_custom_type(PixelParams)


def pixel() -> base.Density2DParameterization:
    """Return the direct pixel parameterization."""

    def from_density_fn(density: types.Density2DArray) -> PixelParams:
        return PixelParams(density=density)

    def to_density_fn(params: PixelParams) -> types.Density2DArray:
        return params.density

    def constraints_fn(params: PixelParams) -> jnp.ndarray:
        del params
        return jnp.asarray(0.0)

    return base.Density2DParameterization(
        from_density=from_density_fn,
        to_density=to_density_fn,
        constraints=constraints_fn,
    )



"""Defines a wrapper for optax optimizers.

Copyright (c) 2023 The INVRS-IO authors.
"""

from typing import Any, Optional, Tuple

import jax
import jax.numpy as jnp
import optax  # type: ignore[import-untyped]
from jax import tree_util
from totypes import types

from invrs_opt.optimizers import base
from invrs_opt.parameterization import (
    base as parameterization_base,
    filter_project,
    gaussian_levelset,
    pixel,
)

PyTree = Any
WrappedOptaxState = Tuple[PyTree, PyTree, PyTree]


def wrapped_optax(opt: optax.GradientTransformation) -> base.Optimizer:
    """Return a wrapped optax optimizer."""
    return parameterized_wrapped_optax(
        opt=opt, penalty=0.0, density_parameterization=None
    )


def density_wrapped_optax(
    opt: optax.GradientTransformation,
    *,
    beta: float,
) -> base.Optimizer:
    """Wrapped optax optimizer with filter-project density parameterization.

    In the filter-project density parameterization, the optimization variable
    associated with a density array is a latent density array; the density is obtained
    by convolving (i.e. "filtering") the latent density with a Gaussian kernel having
    full-width at half-maximum equal to the length scale (the mean of declared minimum
    width and minimum spacing). Then, a tanh nonlinearity is used as a smooth threshold
    operation ("projection").

    Args:
        opt: The optax optimizer to be wrapped.
        beta: Determines the sharpness of the thresholding operation.

    Returns:
        The wrapped optax optimizer.
    """
    return parameterized_wrapped_optax(
        opt=opt,
        penalty=0.0,
        density_parameterization=filter_project.filter_project(beta=beta),
    )


def levelset_wrapped_optax(
    opt: optax.GradientTransformation,
    *,
    penalty: float,
    length_scale_spacing_factor: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_SPACING_FACTOR
    ),
    length_scale_fwhm_factor: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_FWHM_FACTOR
    ),
    length_scale_constraint_factor: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_CONSTRAINT_FACTOR
    ),
    smoothing_factor: int = gaussian_levelset.DEFAULT_SMOOTHING_FACTOR,
    length_scale_constraint_beta: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_CONSTRAINT_BETA
    ),
    length_scale_constraint_weight: float = (
        gaussian_levelset.DEFAULT_LENGTH_SCALE_CONSTRAINT_WEIGHT
    ),
    curvature_constraint_weight: float = (
        gaussian_levelset.DEFAULT_CURVATURE_CONSTRAINT_WEIGHT
    ),
    fixed_pixel_constraint_weight: float = (
        gaussian_levelset.DEFAULT_FIXED_PIXEL_CONSTRAINT_WEIGHT
    ),
    init_optimizer: optax.GradientTransformation = (
        gaussian_levelset.DEFAULT_INIT_OPTIMIZER
    ),
    init_steps: int = gaussian_levelset.DEFAULT_INIT_STEPS,
) -> base.Optimizer:
    """Wrapped optax optimizer with levelset density parameterization.

    In the levelset parameterization, the optimization variable associated with a
    density array is an array giving the amplitudes of Gaussian radial basis functions
    that represent a levelset function over the domain of the density. In the levelset
    parameterization, gradients are nonzero only at the edges of features, and in
    general the topology of a solution does not change during the course of
    optimization.

    The spacing and full-width at half-maximum of the Gaussian basis functions gives
    some amount of control over length scales. In addition, constraints associated with
    length scale, radius of curvature, and deviation from fixed pixels are
    automatically computed and penalized with a weight given by `penalty`. In general,
    this helps ensure that features in an optimized density array violate the specified
    constraints to a lesser degree. The constraints are based on "Analytical level set
    fabrication constraints for inverse design," by D. Vercruysse et al. (2019).

    Args:
        opt: The optax optimizer to be wrapped.
        penalty: The weight of the fabrication penalty, which combines length scale,
            curvature, and fixed pixel constraints.
        length_scale_spacing_factor: The number of levelset control points per unit of
            minimum length scale (mean of density minimum width and minimum spacing).
        length_scale_fwhm_factor: The ratio of Gaussian full-width at half-maximum to
            the minimum length scale.
        length_scale_constraint_factor: Multiplies the target length scale in the
            levelset constraints. A value greater than 1 is pessimistic and drives the
            solution to have a larger length scale (relative to smaller values).
        smoothing_factor: For values greater than 1, the density is initially computed
            at higher resolution and then downsampled, yielding smoother geometries.
        length_scale_constraint_beta: Controls relaxation of the length scale
            constraint near the zero level.
        length_scale_constraint_weight: The weight of the length scale constraint in
            the overall fabrication constraint peenalty.
        curvature_constraint_weight: The weight of the curvature constraint.
        fixed_pixel_constraint_weight: The weight of the fixed pixel constraint.
        init_optimizer: The optimizer used in the initialization of the levelset
            parameterization. At initialization, the latent parameters are optimized so
            that the initial parameters match the binarized initial density.
        init_steps: The number of optimization steps used in the initialization.

    Returns:
        The wrapped optax optimizer.
    """
    return parameterized_wrapped_optax(
        opt=opt,
        penalty=penalty,
        density_parameterization=gaussian_levelset.gaussian_levelset(
            length_scale_spacing_factor=length_scale_spacing_factor,
            length_scale_fwhm_factor=length_scale_fwhm_factor,
            length_scale_constraint_factor=length_scale_constraint_factor,
            smoothing_factor=smoothing_factor,
            length_scale_constraint_beta=length_scale_constraint_beta,
            length_scale_constraint_weight=length_scale_constraint_weight,
            curvature_constraint_weight=curvature_constraint_weight,
            fixed_pixel_constraint_weight=fixed_pixel_constraint_weight,
            init_optimizer=init_optimizer,
            init_steps=init_steps,
        ),
    )


# -----------------------------------------------------------------------------
# Base parameterized wrapped optax optimizer.
# -----------------------------------------------------------------------------


def parameterized_wrapped_optax(
    opt: optax.GradientTransformation,
    density_parameterization: Optional[parameterization_base.Density2DParameterization],
    penalty: float,
) -> base.Optimizer:
    """Wrapped optax optimizer with specified density parameterization.

    Args:
        opt: The optax `GradientTransformation` to be wrapped.
        density_parameterization: The parameterization to be used, or `None`. When no
            parameterization is given, the direct pixel parameterization is used for
            density arrays.
        penalty: The weight of the scalar penalty formed from the constraints of the
            parameterization.

    Returns:
        The `base.Optimizer`.
    """

    if density_parameterization is None:
        density_parameterization = pixel.pixel()

    def _init_latents(params: PyTree) -> PyTree:
        def _leaf_init_latents(leaf: Any) -> Any:
            leaf = _clip(leaf)
            if not _is_density(leaf):
                return leaf
            return density_parameterization.from_density(leaf)

        return tree_util.tree_map(_leaf_init_latents, params, is_leaf=_is_custom_type)

    def _params_from_latents(params: PyTree) -> PyTree:
        def _leaf_params_from_latents(leaf: Any) -> Any:
            if not _is_parameterized_density(leaf):
                return leaf
            return density_parameterization.to_density(leaf)

        return tree_util.tree_map(
            _leaf_params_from_latents,
            params,
            is_leaf=_is_parameterized_density,
        )

    def _constraint_loss(latent_params: PyTree) -> jnp.ndarray:
        def _constraint_loss_leaf(
            params: parameterization_base.ParameterizedDensity2DArrayBase,
        ) -> jnp.ndarray:
            constraints = density_parameterization.constraints(params)
            constraints = tree_util.tree_map(
                lambda x: jnp.sum(jnp.maximum(x, 0.0)),
                constraints,
            )
            return jnp.sum(jnp.asarray(constraints))

        losses = [0.0] + [
            _constraint_loss_leaf(p)
            for p in tree_util.tree_leaves(
                latent_params, is_leaf=_is_parameterized_density
            )
            if _is_parameterized_density(p)
        ]
        return penalty * jnp.sum(jnp.asarray(losses))

    def init_fn(params: PyTree) -> WrappedOptaxState:
        """Initializes the optimization state."""
        latent_params = _init_latents(params)
        params = _params_from_latents(latent_params)
        return params, latent_params, opt.init(latent_params)

    def params_fn(state: WrappedOptaxState) -> PyTree:
        """Returns the parameters for the given `state`."""
        params, _, _ = state
        return params

    def update_fn(
        *,
        grad: PyTree,
        value: float,
        params: PyTree,
        state: WrappedOptaxState,
    ) -> WrappedOptaxState:
        """Updates the state."""
        del value, params

        _, latent_params, opt_state = state
        _, vjp_fn = jax.vjp(_params_from_latents, latent_params)
        (latent_grad,) = vjp_fn(grad)

        if not (
            tree_util.tree_structure(latent_grad)
            == tree_util.tree_structure(latent_params)  # type: ignore[operator]
        ):
            raise ValueError(
                f"Tree structure of `latent_grad` was different than expected, got \n"
                f"{tree_util.tree_structure(latent_grad)} but expected \n"
                f"{tree_util.tree_structure(latent_params)}."
            )

        constraint_loss_grad = jax.grad(_constraint_loss)(latent_params)
        latent_grad = tree_util.tree_map(
            lambda a, b: a + b, latent_grad, constraint_loss_grad
        )

        updates, opt_state = opt.update(
            updates=latent_grad, state=opt_state, params=latent_params
        )
        latent_params = optax.apply_updates(params=latent_params, updates=updates)
        latent_params = _clip(latent_params)
        params = _params_from_latents(latent_params)
        return params, latent_params, opt_state

    return base.Optimizer(init=init_fn, params=params_fn, update=update_fn)


def _is_density(leaf: Any) -> Any:
    """Return `True` if `leaf` is a density array."""
    return isinstance(leaf, types.Density2DArray)


def _is_parameterized_density(leaf: Any) -> Any:
    """Return `True` if `leaf` is a parameterized density array."""
    return isinstance(leaf, parameterization_base.ParameterizedDensity2DArrayBase)


def _is_custom_type(leaf: Any) -> bool:
    """Return `True` if `leaf` is a recognized custom type."""
    return isinstance(leaf, (types.BoundedArray, types.Density2DArray))


def _clip(pytree: PyTree) -> PyTree:
    """Clips leaves on `pytree` to their bounds."""

    def _clip_fn(leaf: Any) -> Any:
        if not _is_custom_type(leaf):
            return leaf
        if leaf.lower_bound is None and leaf.upper_bound is None:
            return leaf
        return tree_util.tree_map(
            lambda x: jnp.clip(x, leaf.lower_bound, leaf.upper_bound), leaf
        )

    return tree_util.tree_map(_clip_fn, pytree, is_leaf=_is_custom_type)



"""Base objects and datatypes for optimization algorithms.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import inspect
from typing import Any, Protocol

import optax  # type: ignore[import-untyped]
from totypes import json_utils

PyTree = Any


class InitFn(Protocol):
    """Callable which initializes an optimizer state."""

    def __call__(self, params: PyTree) -> PyTree:
        ...


class ParamsFn(Protocol):
    """Callable which returns the parameters for an optimizer state."""

    def __call__(self, state: PyTree) -> PyTree:
        ...


class UpdateFn(Protocol):
    """Callable which updates an optimizer state."""

    def __call__(
        self,
        *,
        grad: PyTree,
        value: float,
        params: PyTree,
        state: PyTree,
    ) -> PyTree:
        ...


@dataclasses.dataclass
class Optimizer:
    """Stores the `(init, params, update)` function triple for an optimizer."""

    init: InitFn
    params: ParamsFn
    update: UpdateFn


# Register all optax state types for serialization.
optax_types = {}
for name, obj in inspect.getmembers(optax):
    if name.endswith("State") and isinstance(obj, type):
        optax_types[obj] = True

for obj in optax_types.keys():
    json_utils.register_custom_type(obj)



"""Defines Gaussian radial basis function levelset parameterization.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
from typing import Any, Tuple

import jax
import jax.numpy as jnp
import optax  # type: ignore[import-untyped]
from jax import tree_util
from totypes import json_utils, symmetry, types

from invrs_opt.parameterization import base, transforms

PyTree = Any

DEFAULT_LENGTH_SCALE_SPACING_FACTOR: float = 2.0
DEFAULT_LENGTH_SCALE_FWHM_FACTOR: float = 1.0
DEFAULT_LENGTH_SCALE_CONSTRAINT_FACTOR: float = 1.15
DEFAULT_SMOOTHING_FACTOR: int = 2
DEFAULT_LENGTH_SCALE_CONSTRAINT_BETA: float = 0.333
DEFAULT_LENGTH_SCALE_CONSTRAINT_WEIGHT: float = 1.0
DEFAULT_CURVATURE_CONSTRAINT_WEIGHT: float = 2.0
DEFAULT_FIXED_PIXEL_CONSTRAINT_WEIGHT: float = 10.0
DEFAULT_INIT_STEPS: int = 50
DEFAULT_INIT_OPTIMIZER: optax.GradientTransformation = optax.adam(1e-1)


@dataclasses.dataclass
class GaussianLevelsetParams(base.ParameterizedDensity2DArrayBase):
    """Parameters of a density represented by a Gaussian levelset.

    Attributes:
        amplitude: Array giving the amplitude of the Gaussian basis function at
            levelset control points.
        length_scale_spacing_factor: The number of levelset control points per unit of
            minimum length scale (mean of density minimum width and minimum spacing).
        length_scale_fwhm_factor: The ratio of Gaussian full-width at half-maximum to
            the minimum length scale.
        smoothing_factor: For values greater than 1, the density is initially computed
            at higher resolution and then downsampled, yielding smoother geometries.
        density_shape: Shape of the density array obtained from the parameters.
        density_metadata: Metadata for the density array obtained from the parameters.
    """

    amplitude: jnp.ndarray
    length_scale_spacing_factor: float
    length_scale_fwhm_factor: float
    smoothing_factor: int
    density_shape: Tuple[int, ...]
    density_metadata: base.Density2DMetadata

    def __post_init__(self) -> None:
        self.density_shape = tuple(self.density_shape)

    def example_density(self) -> types.Density2DArray:
        """Returns an example density with appropriate shape and metadata."""
        with jax.ensure_compile_time_eval():
            return types.Density2DArray(
                array=jnp.zeros(self.density_shape),
                **dataclasses.asdict(self.density_metadata),
            )


_GaussianLevelsetParamsAux = Tuple[
    float, float, int, Tuple[int, ...], tree_util.PyTreeDef
]


def _flatten_gaussian_levelset_params(
    params: GaussianLevelsetParams,
) -> Tuple[Tuple[jnp.ndarray], _GaussianLevelsetParamsAux]:
    _, flat_metadata = tree_util.tree_flatten(params.density_metadata)
    return (
        (params.amplitude,),
        (
            params.length_scale_spacing_factor,
            params.length_scale_fwhm_factor,
            params.smoothing_factor,
            params.density_shape,
            flat_metadata,
        ),
    )


def _unflatten_gaussian_levelset_params(
    aux: _GaussianLevelsetParamsAux,
    children: Tuple[jnp.ndarray],
) -> GaussianLevelsetParams:
    (amplitude,) = children
    (
        length_scale_spacing_factor,
        length_scale_fwhm_factor,
        smoothing_factor,
        density_shape,
        flat_metadata,
    ) = aux

    density_metadata = tree_util.tree_unflatten(flat_metadata, ())
    return GaussianLevelsetParams(
        amplitude=amplitude,
        length_scale_spacing_factor=length_scale_spacing_factor,
        length_scale_fwhm_factor=length_scale_fwhm_factor,
        smoothing_factor=smoothing_factor,
        density_shape=tuple(density_shape),
        density_metadata=density_metadata,
    )


tree_util.register_pytree_node(
    GaussianLevelsetParams,
    flatten_func=_flatten_gaussian_levelset_params,
    unflatten_func=_unflatten_gaussian_levelset_params,
)


json_utils.register_custom_type(GaussianLevelsetParams)


def gaussian_levelset(
    *,
    length_scale_spacing_factor: float = DEFAULT_LENGTH_SCALE_SPACING_FACTOR,
    length_scale_fwhm_factor: float = DEFAULT_LENGTH_SCALE_FWHM_FACTOR,
    length_scale_constraint_factor: float = DEFAULT_LENGTH_SCALE_CONSTRAINT_FACTOR,
    smoothing_factor: int = DEFAULT_SMOOTHING_FACTOR,
    length_scale_constraint_beta: float = DEFAULT_LENGTH_SCALE_CONSTRAINT_BETA,
    length_scale_constraint_weight: float = DEFAULT_LENGTH_SCALE_CONSTRAINT_WEIGHT,
    curvature_constraint_weight: float = DEFAULT_CURVATURE_CONSTRAINT_WEIGHT,
    fixed_pixel_constraint_weight: float = DEFAULT_FIXED_PIXEL_CONSTRAINT_WEIGHT,
    init_optimizer: optax.GradientTransformation = DEFAULT_INIT_OPTIMIZER,
    init_steps: int = DEFAULT_INIT_STEPS,
) -> base.Density2DParameterization:
    """Defines a levelset parameterization with Gaussian radial basis functions.

    Args:
        length_scale_spacing_factor: The number of levelset control points per unit of
            minimum length scale (mean of density minimum width and minimum spacing).
        length_scale_fwhm_factor: The ratio of Gaussian full-width at half-maximum to
            the minimum length scale.
        length_scale_constraint_factor: Multiplies the target length scale in the
            levelset constraints. A value greater than 1 is pessimistic and drives the
            solution to have a larger length scale (relative to smaller values).
        smoothing_factor: For values greater than 1, the density is initially computed
            at higher resolution and then downsampled, yielding smoother geometries.
        length_scale_constraint_beta: Controls relaxation of the length scale
            constraint near the zero level.
        length_scale_constraint_weight: The weight of the length scale constraint in
            the overall fabrication constraint peenalty.
        curvature_constraint_weight: The weight of the curvature constraint.
        fixed_pixel_constraint_weight: The weight of the fixed pixel constraint.
        init_optimizer: The optimizer used in the initialization of the levelset
            parameterization. At initialization, the latent parameters are optimized so
            that the initial parameters match the binarized initial density.
        init_steps: The number of optimization steps used in the initialization.

    Returns:
        The `Density2DParameterization`.
    """

    def from_density_fn(density: types.Density2DArray) -> GaussianLevelsetParams:
        """Return level set parameters for the given `density`."""
        length_scale = (density.minimum_width + density.minimum_spacing) / 2
        spacing_factor = length_scale_spacing_factor / length_scale
        shape = density.shape[:-2] + (
            int(jnp.ceil(density.shape[-2] * spacing_factor)),
            int(jnp.ceil(density.shape[-1] * spacing_factor)),
        )

        mid_value = 0.5 * (density.lower_bound + density.upper_bound)
        value_range = density.upper_bound - density.lower_bound
        target_array = transforms.apply_fixed_pixels(density).array
        target_array = (
            jnp.sign(target_array - mid_value) * 0.5 * value_range + mid_value
        )

        # Generate the initial amplitude array.
        amplitude = density.array - (density.upper_bound + density.lower_bound) / 2
        amplitude = jnp.sign(amplitude)
        amplitude = transforms.resample(amplitude, shape)

        # If the density is not periodic, ensure there are level set control points
        # beyond the edge of the density array.
        pad_width = ((0, 0),) * (amplitude.ndim - 2)
        pad_width += ((0, 0),) if density.periodic[0] else ((1, 1),)
        pad_width += ((0, 0),) if density.periodic[1] else ((1, 1),)
        amplitude = jnp.pad(amplitude, pad_width, mode="edge")

        density_metadata_dict = dataclasses.asdict(density)
        del density_metadata_dict["array"]
        density_metadata = base.Density2DMetadata(**density_metadata_dict)
        params = GaussianLevelsetParams(
            amplitude=amplitude,
            length_scale_spacing_factor=length_scale_spacing_factor,
            length_scale_fwhm_factor=length_scale_fwhm_factor,
            smoothing_factor=smoothing_factor,
            density_shape=density.shape,
            density_metadata=density_metadata,
        )

        def step_fn(
            _: int,
            params_and_state: Tuple[PyTree, PyTree],
        ) -> Tuple[PyTree, PyTree]:
            def loss_fn(params: GaussianLevelsetParams) -> jnp.ndarray:
                density_from_params = to_density_fn(params, mask_gradient=False)
                return jnp.mean((density_from_params.array - target_array) ** 2)

            params, state = params_and_state
            grad = jax.grad(loss_fn)(params)
            updates, state = init_optimizer.update(grad, params=params, state=state)
            params = optax.apply_updates(params, updates)
            return params, state

        state = init_optimizer.init(params)
        params, state = jax.lax.fori_loop(
            0, init_steps, body_fun=step_fn, init_val=(params, state)
        )

        maxval = jnp.amax(jnp.abs(params.amplitude), axis=(-2, -1), keepdims=True)
        return dataclasses.replace(params, amplitude=params.amplitude / maxval)

    def to_density_fn(
        params: GaussianLevelsetParams,
        mask_gradient: bool = True,
    ) -> types.Density2DArray:
        """Return a density from the latent parameters."""
        array = _to_array(params, mask_gradient=mask_gradient, pad_pixels=0)

        example_density = params.example_density()
        lb = example_density.lower_bound
        ub = example_density.upper_bound
        array = lb + array * (ub - lb)
        assert array.shape == example_density.shape
        return dataclasses.replace(example_density, array=array)

    def constraints_fn(
        params: GaussianLevelsetParams,
        mask_gradient: bool = True,
        pad_pixels: int = 2,
    ) -> jnp.ndarray:
        """Computes constraints associated with the params."""
        length_scale_constraint, curvature_constraint = _levelset_constraints(
            params,
            beta=length_scale_constraint_beta,
            length_scale_constraint_factor=length_scale_constraint_factor,
            pad_pixels=pad_pixels,
        )
        fixed_pixel_constraint = _fixed_pixel_constraint(
            params,
            mask_gradient=mask_gradient,
            pad_pixels=pad_pixels,
        )

        constraints = jnp.stack(
            [
                length_scale_constraint * length_scale_constraint_weight,
                curvature_constraint * curvature_constraint_weight,
                fixed_pixel_constraint * fixed_pixel_constraint_weight,
            ],
            axis=-1,
        )

        # Normalize constraints to make them (somewhat) resolution-independent.
        example_density = params.example_density()
        length_scale = 0.5 * (
            example_density.minimum_spacing + example_density.minimum_width
        )
        return constraints / length_scale**2

    return base.Density2DParameterization(
        to_density=to_density_fn,
        from_density=from_density_fn,
        constraints=constraints_fn,
    )


# -----------------------------------------------------------------------------
# Functions to obtain arrays from the levelset parameterization.
# -----------------------------------------------------------------------------


def _to_array(
    params: GaussianLevelsetParams,
    mask_gradient: bool,
    pad_pixels: int,
) -> jnp.ndarray:
    """Return an array from the parameters.

    The array has a value of `1` where the levelset array is positive, and a value
    of `-1` elsewhere. The final density array can be obtained by rescaling this array
    to have the appropriate upper and lower bounds.

    Args:
        params: The parameters from which the density is obtained.
        mask_gradient: If `True`, the gradient is masked so that it is nonzero only at
            the borders of features.
        pad_pixels: A non-negative integer giving the additional pixels to be included
            beyond the boundaries of the parameterized density.

    Returns:
        The array.
    """
    example_density = params.example_density()
    periodic: Tuple[bool, bool] = example_density.periodic
    phi = _phi_from_params(
        params=params,
        pad_pixels=pad_pixels,
    )
    array = _levelset_threshold(
        phi=phi,
        periodic=periodic,
        mask_gradient=mask_gradient,
    )
    return _downsample_spatial_dims(array, params.smoothing_factor)


def _phi_from_params(
    params: GaussianLevelsetParams,
    pad_pixels: int,
) -> jnp.ndarray:
    """Return the levelset function for the given `params`.

    Args:
        params: The parameters from which the density is obtained.
        pad_pixels: A non-negative integer giving the additional pixels to be included
            beyond the boundaries of the parameterized density.

    Returns:
        The levelset array `phi`.
    """
    with jax.ensure_compile_time_eval():
        example_density = params.example_density()
        length_scale = 0.5 * (
            example_density.minimum_width + example_density.minimum_spacing
        )
        fwhm = length_scale * params.length_scale_fwhm_factor
        sigma = fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))

        highres_i = (
            0.5
            + jnp.arange(
                params.smoothing_factor * (-pad_pixels),
                params.smoothing_factor * (pad_pixels + example_density.shape[-2]),
            )
        ) / params.smoothing_factor
        highres_j = (
            0.5
            + jnp.arange(
                params.smoothing_factor * (-pad_pixels),
                params.smoothing_factor * (pad_pixels + example_density.shape[-1]),
            )
        ) / params.smoothing_factor

        # Coordinates for the control points of the Gaussian radial basis functions.
        levelset_i, levelset_j = _control_point_coords(
            density_shape=params.density_shape[-2:],  # type: ignore[arg-type]
            levelset_shape=params.amplitude.shape[-2:],  # type: ignore[arg-type]
            periodic=example_density.periodic,
        )

        # Handle periodicity by replicating control points over a 3x3 supercell.
        if example_density.periodic[0]:
            levelset_i = jnp.concatenate(
                [
                    levelset_i - example_density.shape[-2],
                    levelset_i,
                    levelset_i + example_density.shape[-2],
                ],
                axis=-2,
            )
            levelset_j = jnp.concatenate([levelset_j] * 3, axis=-2)
        if example_density.periodic[1]:
            levelset_i = jnp.concatenate([levelset_i] * 3, axis=-1)
            levelset_j = jnp.concatenate(
                [
                    levelset_j - example_density.shape[-1],
                    levelset_j,
                    levelset_j + example_density.shape[-1],
                ],
                axis=-1,
            )

        levelset_i = levelset_i.flatten()
        levelset_j = levelset_j.flatten()

    amplitude = params.amplitude
    if example_density.periodic[0]:
        amplitude = jnp.concat([amplitude] * 3, axis=-2)
    if example_density.periodic[1]:
        amplitude = jnp.concat([amplitude] * 3, axis=-1)

    amplitude = amplitude.reshape(amplitude.shape[:-2] + (1, -1))

    # Use a scan operation to compute the array; this lowers memory consumption.
    def scan_fn(_: Tuple[()], i: jnp.ndarray) -> Tuple[Tuple[()], jnp.ndarray]:
        distance_sq = (i - levelset_i) ** 2 + (
            highres_j[:, jnp.newaxis] - levelset_j
        ) ** 2
        basis = jnp.exp(-distance_sq / sigma**2)
        return (), jnp.sum(basis * amplitude, axis=-1)

    _, array = jax.lax.scan(scan_fn, (), xs=highres_i)
    array = jnp.moveaxis(array, 0, -2)

    assert array.shape[-2] % params.smoothing_factor == 0
    assert array.shape[-1] % params.smoothing_factor == 0
    array = symmetry.symmetrize(array, tuple(example_density.symmetries))
    return array


# -----------------------------------------------------------------------------
# Functions to compute constraints.
# -----------------------------------------------------------------------------


def _fixed_pixel_constraint(
    params: GaussianLevelsetParams,
    mask_gradient: bool,
    pad_pixels: int,
) -> jnp.ndarray:
    """Return the fixed pixel constraint array.

    The fixed pixel constraint array is nonzero at locations where the density obtained
    from `params` differs from fixed pixels.

    Args:
        params: The parameters from which the density is obtained.
        mask_gradient: If `True`, the gradient is masked so that it is nonzero only at
            the borders of features.
        pad_pixels: The number of pixels added at borders. Values greater than zero
            help to ensure that sharp features at the borders are avoided.

    Returns:
        The constraints array.
    """
    array = _to_array(params, mask_gradient=mask_gradient, pad_pixels=pad_pixels)

    example_density = params.example_density()
    fixed_solid = jnp.zeros(example_density.shape, dtype=bool)
    fixed_void = jnp.zeros(example_density.shape, dtype=bool)
    if example_density.fixed_solid is not None:
        fixed_solid = jnp.asarray(example_density.fixed_solid)
    if example_density.fixed_void is not None:
        fixed_void = jnp.asarray(example_density.fixed_void)

    pad_width = ((0, 0),) * (fixed_solid.ndim - 2) + (
        (pad_pixels, pad_pixels),
        (pad_pixels, pad_pixels),
    )
    fixed_solid = jnp.pad(fixed_solid, pad_width, mode="edge")
    fixed_void = jnp.pad(fixed_void, pad_width, mode="edge")
    fixed = fixed_solid | fixed_void
    target = jnp.where(fixed_solid, 1, 0)

    return jnp.where(fixed, jnp.abs(array - target), 0.0)


def _levelset_constraints(
    params: GaussianLevelsetParams,
    beta: float,
    length_scale_constraint_factor: float,
    pad_pixels: int,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute constraints for minimum width, spacing, and radius of curvature.

    The constraints are based on "Analytical level set fabrication constraints for
    inverse design," by D. Vercruysse et al. (2019). Constraints are satisfied when
    they are non-positive.

    https://www.nature.com/articles/s41598-019-45026-0

    Args:
        params: The parameters of the Gaussian levelset.
        beta: Parameter which relaxes the constraint near the zero-plane.
        length_scale_constraint_factor: Multiplies the target length scale in the
            levelset constraints. A value greater than 1 is pessimistic and drives the
            solution to have a larger length scale (relative to smaller values).
        pad_pixels: A non-negative integer giving the additional pixels to be included
            beyond the boundaries of the parameterized density.

    Returns:
        The minimum length scale and minimum curvature constraint arrays.
    """
    example_density = params.example_density()
    minimum_length_scale = 0.5 * (
        example_density.minimum_width + example_density.minimum_spacing
    )

    phi, phi_v, phi_vv, inverse_radius = _phi_derivatives_and_inverse_radius(
        params,
        pad_pixels=pad_pixels,
    )

    d = minimum_length_scale * length_scale_constraint_factor
    length_scale_constraint = (
        jnp.abs(phi_vv) / (jnp.pi / d * jnp.abs(phi) + beta * phi_v) - jnp.pi / d
    )
    curvature_constraint = (
        jnp.abs(inverse_radius * jnp.arctan(phi_v / phi)) - jnp.pi / d
    )

    # Downsample so that constraints shape matches the density shape.
    return (
        _downsample_spatial_dims(length_scale_constraint, params.smoothing_factor),
        _downsample_spatial_dims(curvature_constraint, params.smoothing_factor),
    )


def _phi_derivatives_and_inverse_radius(
    params: GaussianLevelsetParams,
    pad_pixels: int,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Compute the levelset function and its first and second derivatives."""

    phi = _phi_from_params(
        params=params,
        pad_pixels=pad_pixels,
    )

    d = 1 / params.smoothing_factor
    phi_x, phi_y = jnp.gradient(phi, d, axis=(-2, -1))
    phi_xx, phi_yx = jnp.gradient(phi_x, d, axis=(-2, -1))
    phi_xy, phi_yy = jnp.gradient(phi_y, d, axis=(-2, -1))

    phi_v = _sqrt_safe(phi_x**2 + phi_y**2)

    # Compute "safe" versions of `phi_v` and its square, which are used to
    # normalize quantities below. These are equal to 1 anywhere `phi_v` is
    # close to zero, and take their usual values elsewhere.
    phi_v_near_zero = jnp.isclose(phi_v, 0.0)
    phi_v_squared_safe = jnp.where(phi_v_near_zero, 1.0, phi_v**2)
    phi_v_safe = jnp.where(phi_v_near_zero, 1.0, phi_v)

    weight_xx = phi_x**2 / phi_v_squared_safe
    weight_yy = phi_y**2 / phi_v_squared_safe
    weight_xy = (phi_x * phi_y) / phi_v_squared_safe
    phi_vv = weight_xx * phi_xx + weight_xy * (phi_xy + phi_yx) + weight_yy * phi_yy

    qx: jnp.ndarray = jnp.gradient(  # type: ignore[assignment]
        phi_x / jnp.abs(phi_v_safe), d, axis=-2
    )
    qy: jnp.ndarray = jnp.gradient(  # type: ignore[assignment]
        phi_y / jnp.abs(phi_v_safe), d, axis=-1
    )
    inverse_radius = qx + qy

    return phi, phi_v, phi_vv, inverse_radius


# -----------------------------------------------------------------------------
# Helper functions.
# -----------------------------------------------------------------------------


def _control_point_coords(
    density_shape: Tuple[int, int],
    levelset_shape: Tuple[int, int],
    periodic: Tuple[bool, bool],
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Returns the control point coordinates."""
    # If the levelset is periodic along any axis, the first and last control
    # points along that axis lie outside the bounds of the density.
    offset_i = 0.5 if periodic[0] else -0.5
    offset_j = 0.5 if periodic[1] else -0.5
    range_i = levelset_shape[-2] - (0 if periodic[0] else 2)
    range_j = levelset_shape[-1] - (0 if periodic[1] else 2)

    factor_i = density_shape[-2] / range_i
    factor_j = density_shape[-1] / range_j
    levelset_i, levelset_j = jnp.meshgrid(
        (offset_i + jnp.arange(levelset_shape[-2])) * factor_i,
        (offset_j + jnp.arange(levelset_shape[-1])) * factor_j,
        indexing="ij",
    )
    return levelset_i, levelset_j


def _sqrt_safe(x: jnp.ndarray) -> jnp.ndarray:
    """Compute square root while avoiding `nan` gradients near zero."""
    x_near_zero = jnp.isclose(x, 0.0)
    x_safe = jnp.where(x_near_zero, 1, x)
    return jnp.where(x_near_zero, 0.0, jnp.sqrt(x_safe))


def _levelset_threshold(
    phi: jnp.ndarray,
    periodic: Tuple[bool, bool],
    mask_gradient: bool,
) -> jnp.ndarray:
    """Thresholds a level set function `phi`."""
    if mask_gradient:
        interface = _interface_pixels(phi, periodic)
        phi = jnp.where(interface, phi, jax.lax.stop_gradient(phi))
    thresholded = (phi > 0).astype(float) + (phi - jax.lax.stop_gradient(phi))
    return thresholded


def _interface_pixels(phi: jnp.ndarray, periodic: Tuple[bool, bool]) -> jnp.ndarray:
    """Identifies interface pixels of a level set function `phi`."""
    batch_shape = phi.shape[:-2]
    phi = phi.reshape((-1,) + phi.shape[-2:])

    pad_mode = (
        "wrap" if periodic[0] else "edge",
        "wrap" if periodic[1] else "edge",
    )
    pad_width = ((1, 1), (1, 1))

    kernel = jnp.asarray([[0, 1, 0], [1, 0, 1], [0, 1, 0]], dtype=float)

    solid = phi > 0
    void = ~solid

    solid_padded = transforms.pad2d(solid, pad_width, pad_mode)
    num_solid_adjacent = transforms.conv(
        x=solid_padded[:, jnp.newaxis, :, :].astype(float),
        kernel=kernel[jnp.newaxis, jnp.newaxis, :, :],
        padding="VALID",
    )
    num_solid_adjacent = jnp.squeeze(num_solid_adjacent, axis=1)

    void_padded = transforms.pad2d(void, pad_width, pad_mode)
    num_void_adjacent = transforms.conv(
        x=void_padded[:, jnp.newaxis, :, :].astype(float),
        kernel=kernel[jnp.newaxis, jnp.newaxis, :, :],
        padding="VALID",
    )
    num_void_adjacent = jnp.squeeze(num_void_adjacent, axis=1)

    interface = solid & (num_void_adjacent > 0) | void & (num_solid_adjacent > 0)

    return interface.reshape(batch_shape + interface.shape[-2:])


def _downsample_spatial_dims(x: jnp.ndarray, downsample_factor: int) -> jnp.ndarray:
    """Downsamples the two trailing axes of `x` by `downsample_factor`."""
    shape = x.shape[:-2] + (
        x.shape[-2] // downsample_factor,
        x.shape[-1] // downsample_factor,
    )
    return transforms.box_downsample(x, shape)
